/* This file is part of the Imote2 Codeloader.
 *
 * The Imote2 Codeloader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Imote2 Codeloader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Imote2 Codeloader. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * Imote.java
 *
 * @author Markus Engel
 * Copyright (c) 2009,2010 Networked Systems Group, University of Kaiserslautern
 */
package imote.communication;

import imote.ImoteSerial;
import imote.service.ImoteInterface;
import ch.ntb.usb.*;
import imote.ImoteDescription;
import java.util.LinkedList;
import java.util.List;

/** Imote 2 Node
 * Protocol implementation for bootloader on top of USB
 */
public class Imote implements ImoteInterface {
	/** USB Configuration number */
	private static final int USB_CONFIGURATION = 1;
	/** USB Interface number */
	private static final int USB_INTERFACE = 0;
	/** USB Alternative setting number */
	private static final int USB_ALT_SETTING = -1;
	/** USB Transmit Endpoint */
	private static final int ENDPOINT_TX = 1;
	/** USB Receive Endpoint */
	private static final int ENDPOINT_RX = 0x82;
	/** USB Device */
	private Device device;
	/** Bootloader Version, encoded in BCD */
	private short deviceVersion;

	/** create new Imote driver from device
	 *
	 * @param device USB Device
	 */
	private Imote(Device device) {
		this.device = device;
		this.deviceVersion = this.device.getDeviceDescriptor().getBcdDevice();
	}

	/** helper to open device by vendor and product ID and optionally serial
	 * number and / or node ID
	 *
	 * @param vendorId vendor id
	 * @param productId product id
	 * @param serial serial number and node id request
	 * @return USB Device
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	private static Device openDevice(int vendorId, int productId,
		ImoteSerial serial) throws USBException {
		/* iterate over USB Busses */
		for (Usb_Bus bus = USB.getBus(); bus != null; bus = bus.getNext()) {
			String busDirname = bus.getDirname();
			/* iterate over devices of that bus */
			for (Usb_Device device = bus.getDevices(); device != null;
				device = device.getNext()) {
				Usb_Device_Descriptor desc = device.getDescriptor();
				if (desc.getIdVendor() == vendorId &&
					desc.getIdProduct() == productId) {
					/* found an imote */
					String devFilename = device.getFilename();
					/* get serial number and node id */
					ImoteSerial ser = ImoteSerial.read(device, desc);
					if (ser.fulfilsRequest(serial)) {
						/* found requested imote => open */
						return USB.getDevice((short)vendorId, (short)productId,
							busDirname, devFilename);
					}
				}
			}
		}

		/* not found */
		return null;
	}

	/** open imote by vendor and product id and optionally serial number
	 * and/or node id
	 *
	 * @param vendorId vendor id
	 * @param productId product id
	 * @param serialNumber serial number or null
	 * @param nodeId node id or null
	 * @return device or null if not found
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public static Imote getImote(int vendorId, int productId,
		String serialNumber, Long nodeId) throws USBException {
		Device dev;
		if (serialNumber == null && nodeId == null)
			dev = USB.getDevice((short)vendorId, (short)productId);
		else
			dev = openDevice(vendorId, productId, new ImoteSerial(serialNumber,
				nodeId));

		if (dev == null)
			return null;

		try {
			dev.setResetOnFirstOpen(true, 100);
			dev.open(USB_CONFIGURATION, USB_INTERFACE, USB_ALT_SETTING);
			return new Imote(dev);
		} catch (USBException e) {
			if (!e.toString().contains("USB device with idVendor")) {
				throw e;
			}
		}

		return null;
	}

	/** get a list of connected imotes
	 *
	 * @param vendorId vendor id
	 * @param productId product id
	 * @return list of imote descriptions
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public static List<ImoteDescription> getConnectedImotes(int vendorId,
		int productId) throws USBException {
		LinkedList<ImoteDescription> l = new LinkedList<ImoteDescription>();

		/* iterate over USB busses */
		for (Usb_Bus bus = USB.getBus(); bus != null; bus = bus.getNext()) {
			/* iterate over devices */
			for (Usb_Device device = bus.getDevices(); device != null;
				device = device.getNext()) {
				Usb_Device_Descriptor desc = device.getDescriptor();
				if (desc.getIdVendor() == vendorId &&
					desc.getIdProduct() == productId) {
					/* found imote, add to list */
					ImoteSerial ser = ImoteSerial.read(device, desc);
					l.add(new ImoteDescription(ser,
						bcdDeviceToString(desc.getBcdDevice())));
				}
			}
		}

		return l;
	}

	/** close device
	 * @note do not use that afterwards
	 */
	public void close() {
		try {
			this.device.close();
		} catch (USBException e) {
		}
	}

	/** get serial number and node id of imote
	 *
	 * @return serial number and node id
	 */
	public ImoteSerial getSerial() {
		return ImoteSerial.read(this.device.getDevice(),
			this.device.getDeviceDescriptor());
	}

	/** get bootloader version as BCD
	 *
	 * @return bootloader version, BCD encoded
	 */
	public short getBcdDeviceVersion() {
		return this.deviceVersion;
	}

	/** helper to decode BCD encoded integer
	 *
	 * @param i BCD encoded integer
	 * @return i as integer
	 */
	private static int bcdDecode(int i) {
		return (i / 16 * 10) + (i % 16);
	}

	/** helper to get bootloader version from BCD encoded version number
	 *
	 * @param bcdVersion BCD encoded version
	 * @return version as readable string
	 */
	private static String bcdDeviceToString(short bcdVersion) {
		return new Integer(bcdDecode(bcdVersion >> 8)).toString() + "." +
			new Integer(bcdDecode(bcdVersion & 0xff)).toString();
	}

	/** get device version as readable string
	 *
	 * @return device version
	 */
	public String getDeviceVersion() {
		return bcdDeviceToString(this.deviceVersion);
	}

	/** send a packet over USB to device
	 *
	 * @param pkt packet to be sent
	 * @return number of bytes sent
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	private int sendPacket(Packet pkt) throws USBException {
		return pkt.send(this.device, ENDPOINT_TX, 100 * 1000);
	}

	/** receive status packet from imote
	 *
	 * @return status packet (only if status is OK)
	 * @throws ImoteException if receiving failed or status is not OK
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	private imote.communication.rx.StatusPacket getStatus() throws
		ImoteException, USBException {
		imote.communication.rx.StatusPacket status =
			new imote.communication.rx.StatusPacket();
		int recvLength;

		recvLength =
			status.recv(this.device, ENDPOINT_RX, 500 * 32 * 1024 / 128);
		if (recvLength != 4) {
			throw new ImoteException(
				imote.communication.rx.Status.ERR_COMMUNICATION_FAILURE);
		}
		if (!status.statusIsOk()) {
			throw new ImoteException(status.getStatus());
		}
		return status;
	}

	/** send RESET message and wait for ACK
	 *
	 * @throws ImoteException if receiving failed or status is not OK
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public void reset() throws ImoteException, USBException {
		this.sendPacket(new imote.communication.tx.Reset());
		this.getStatus();
	}

	/** send REBOOT message and wait for ACK
	 *
	 * @throws ImoteException if receiving failed or status is not OK
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public void reboot() throws ImoteException, USBException {
		this.sendPacket(new imote.communication.tx.Reboot());
		try {
			this.getStatus();
		} catch (ImoteException e) {
			if (e.getStatus()
				!= imote.communication.rx.Status.ERR_COMMUNICATION_FAILURE) {
				throw e;
			}
		} catch(USBException u) {}
	}

	/** send ADDRESS message and wait for ACK
	 *
	 * @param address address which should be programmed next
	 * @param length length of the programming request
	 * @throws ImoteException if receiving failed or status is not OK
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public void address(long address, long length) throws ImoteException,
		USBException {
		this.sendPacket(new imote.communication.tx.Address(address, length));
		this.getStatus();
	}

	/** send ERASE message and wait for ACK
	 *
	 * @throws ImoteException if receiving failed or status is not OK
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public void erase() throws ImoteException, USBException {
		this.sendPacket(new imote.communication.tx.Erase());
		this.getStatus();
	}

	/** send UPLOAD message and wait for ACK
	 *
	 * @param data data to be sent in the message
	 * @param offset offset in data where transmit begins
	 * @param length length of data to be sent
	 * @param compressed flag wether data are compressed
	 * @return number of bytes sent
	 * @throws ImoteException if receiving failed or status is not OK
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public int upload(byte[] data, int offset, int length, boolean compressed)
		throws ImoteException, USBException {
		int ret = this.sendPacket(
			new imote.communication.tx.Upload(data, offset, length, compressed));
		this.getStatus();
		return ret;
	}

	/** send UPLOAD_FINISHED message and wait for ACK
	 *
	 * @return number of bytes sent
	 * @throws ImoteException if receiving failed or status is not OK
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public int uploadFinish() throws ImoteException, USBException {
		int ret = this.sendPacket(new imote.communication.tx.UploadFinish());
		this.getStatus();
		return ret;
	}

	/** send EXECUTE_RAM message
	 * @note does not wait for ACK, because this will not send any
	 * @return number of bytes sent
	 * @throws ImoteException if receiving failed or status is not OK
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public int executeRam() throws ImoteException, USBException {
		int ret = this.sendPacket(new imote.communication.tx.ExecuteRam());
		//this.getStatus();
		return ret;
	}
}
